import json
import urllib
import urllib2
import urlparse


ENDPOINT = None
KEY = None


class Scout(object):
    def __init__(self, endpoint=ENDPOINT, key=KEY):
        self.endpoint = endpoint
        self.key = key

    def get_full_url(self, url):
        return urlparse.urljoin(self.endpoint, url)

    def get(self, url, **kwargs):
        headers = {'Content-Type': 'application/json'}
        if self.key:
            headers['key'] = self.key
        if kwargs:
            if '?' not in url:
                url += '?'
            url += urllib.urlencode(kwargs, True)
        request = urllib2.Request(self.get_full_url(url), headers=headers)
        fh = urllib2.urlopen(request)
        return json.loads(fh.read())

    def post(self, url, data=None):
        headers = {'Content-Type': 'application/json'}
        if self.key:
            headers['key'] = self.key
        request = urllib2.Request(
            self.get_full_url(url),
            data=json.dumps(data or {}),
            headers=headers)
        fh = urllib2.urlopen(request)
        return json.loads(fh.read())

    def delete(self, url):
        headers = {}
        if self.key:
            headers['key'] = self.key
        request = urllib2.Request(self.get_full_url(url), headers=headers)
        request.get_method = lambda: 'DELETE'
        fh = urllib2.urlopen(request)
        return json.loads(fh.read())

    def get_indexes(self):
        return self.get('/')['indexes']

    def create_index(self, name):
        return self.post('/', {'name': name})

    def rename_index(self, old_name, new_name):
        return self.post('/%s/' % old_name, {'name': new_name})

    def delete_index(self, name):
        return self.delete('/%s/' % name)

    def get_documents(self, **kwargs):
        return self.get('/documents/', **kwargs)

    def store_document(self, content, indexes, identifier=None, **metadata):
        if not isinstance(indexes, (list, tuple)):
            indexes = [indexes]
        return self.post('/documents/', {
            'content': content,
            'identifier': identifier,
            'indexes': indexes,
            'metadata': metadata})

    def update_document(self, document_id=None, content=None, indexes=None,
                        metadata=None, identifier=None):
        if not document_id and not identifier:
            raise ValueError('`document_id` or `identifier` must be provided.')

        data = {}
        if content is not None:
            data['content'] = content
        if indexes is not None:
            if not isinstance(indexes, (list, tuple)):
                indexes = [indexes]
            data['indexes'] = indexes
        if metadata is not None:
            data['metadata'] = metadata

        if not data:
            raise ValueError('Nothing to update.')

        if document_id:
            return self.post('/documents/%s/' % document_id, data)
        else:
            return self.post('/documents/identifier/%s/' % identifier, data)

    def delete_document(self, document_id=None, identifier=None):
        if not document_id and not identifier:
            raise ValueError('`document_id` or `identifier` must be provided.')

        if document_id:
            return self.delete('/documents/%s/' % document_id)
        else:
            return self.delete('/documents/identifier/%s/' % identifier)

    def search(self, index, query, **kwargs):
        kwargs['q'] = query
        return self.get('/%s/search/' % index, **kwargs)



#!/usr/bin/env python

"""
RESTful search server powered by SQLite's full-text search extension.
"""
__author__ = 'Charles Leifer'
__kitty__ = 'Huey'
__version__ = '0.2.1'

try:
    from functools import reduce
except ImportError:
    pass
from functools import wraps
import json
import operator
import optparse
import os
import sys

from flask import abort, Flask, jsonify, request, Response
from peewee import *
from playhouse.flask_utils import get_object_or_404
from playhouse.flask_utils import PaginatedQuery
from playhouse.sqlite_ext import *
from playhouse.sqlite_ext import _VirtualFieldMixin
from werkzeug.serving import run_simple


AUTHENTICATION = None
DATABASE = None
DEBUG = False
HOST = '127.0.0.1'
PAGE_VAR = 'page'
PAGINATE_BY = 50
PORT = 8000
SECRET_KEY = 'huey is a little angel.'  # Customize this.
STEM = 'porter'

app = Flask(__name__)
app.config.from_object(__name__)

database = SqliteExtDatabase(None)

#
# Database models.
#

class Document(FTSModel):
    """
    The :py:class:`Document` class contains content which should be indexed
    for search. Documents can be associated with any number of indexes via
    the `IndexDocument` junction table. Because `Document` is implemented
    as an FTS3 virtual table, it does not support any secondary indexes, and
    all columns have *Text* type, regardless of their declared type. For that
    reason we will utilize the internal SQLite `rowid` column to relate
    documents to indexes.
    """
    rowid = RowIDField()
    content = TextField()
    identifier = TextField()

    class Meta:
        database = database
        db_table = 'main_document'

    @classmethod
    def all(cls):
        # Explicitly select the `rowid`, otherwise it would not be selected.
        return Document.select(
            Document.rowid,
            Document.content,
            Document.identifier)

    def get_metadata(self):
        return dict(Metadata
                    .select(Metadata.key, Metadata.value)
                    .where(Metadata.document == self.rowid)
                    .tuples())

    def set_metadata(self, metadata):
        (Metadata
         .insert_many([
             {'key': key, 'value': value, 'document': self.rowid}
             for key, value in metadata.items()])
         .execute())

    def delete_metadata(self):
        Metadata.delete().where(Metadata.document == self.rowid).execute()

    metadata = property(get_metadata, set_metadata, delete_metadata)

    def get_indexes(self):
        return (Index
                .select()
                .join(IndexDocument)
                .where(IndexDocument.document == self))


class BaseModel(Model):
    class Meta:
        database = database


class Metadata(BaseModel):
    """
    Arbitrary key/value pairs associated with an indexed `Document`. The
    metadata associated with a document can also be used to filter search
    results.
    """
    document = ForeignKeyField(Document, related_name='metadata_set')
    key = CharField()
    value = TextField()

    class Meta:
        db_table = 'main_metadata'
        indexes = (
            (('document', 'key'), True),
        )


class Index(BaseModel):
    """
    Indexes contain any number of documents and expose a clean API for
    searching and storing content.
    """
    RANK_SIMPLE = 'simple'
    RANK_BM25 = 'bm25'
    RANK_NONE = None

    name = CharField(unique=True)

    class Meta:
        db_table = 'main_index'

    def search(self, search, ranking=RANK_SIMPLE, **filters):
        if ranking == Index.RANK_SIMPLE:
            rank_expr = Document.rank()
        elif ranking == Index.RANK_BM25:
            rank_expr = Document.bm25(Document.content)

        selection = [Document.rowid, Document.content, Document.identifier]
        if ranking:
            selection.append(rank_expr.alias('score'))

        query = (Document
                 .select(*selection)
                 .join(IndexDocument)
                 .where(
                     (IndexDocument.index == self) &
                     (Document.match(search))))

        if filters:
            filter_expr = reduce(operator.and_, [
                fn.EXISTS(Metadata.select().where(
                    (Metadata.key == key) &
                    (Metadata.value == value) &
                    (Metadata.document == Document.rowid)))
                for key, value in filters.items()
            ])
            query = query.where(filter_expr)

        if ranking:
            query = query.order_by(rank_expr.desc())

        return query

    def add_to_index(self, document):
        with database.atomic():
            try:
                IndexDocument.create(index=self, document=document)
            except IntegrityError:
                pass

    def index(self, content, document=None, identifier=None, **metadata):
        if document is None:
            document = Document()
        else:
            del document.metadata
        document.content = content
        document.identifier = identifier
        document.save()
        self.add_to_index(document)

        if metadata:
            document.metadata = metadata
        return document

    @property
    def documents(self):
        return (Document
                .all()
                .join(IndexDocument)
                .where(IndexDocument.index == self))


class IndexDocument(BaseModel):
    index = ForeignKeyField(Index)
    document = ForeignKeyField(Document)

    class Meta:
        db_table = 'main_index_document'
        indexes = (
            (('index', 'document'), True),
        )

#
# View helpers.
#

def parse_post(required_keys=None, optional_keys=None):
    """
    Clean and validate POSTed JSON data by defining sets of required and
    optional keys.
    """
    try:
        data = json.loads(request.data)
    except ValueError:
        error('Unable to parse JSON data from request.')

    required = set(required_keys or [])
    optional = set(optional_keys or [])
    all_keys = required | optional
    keys_present = set(key for key in data if data[key] not in ('', None))

    missing = required - keys_present
    if missing:
        error('Missing required fields: %s' % ', '.join(sorted(missing)))

    invalid_keys = keys_present - all_keys
    if invalid_keys:
        error('Invalid keys: %s' % ', '.join(sorted(invalid_keys)))

    return data

class InvalidRequestException(Exception):
    def __init__(self, error_message):
        self.error_message = error_message

    def response(self):
        return jsonify({'error': self.error_message}), 400

def error(message):
    """
    Trigger an Exception from a view that will short-circuit the Response
    cycle and return a 400 "Bad request" with the given error message.
    """
    raise InvalidRequestException(message)

def validate_indexes(data, required=True):
    if data.get('index'):
        index_names = [data['index']]
    elif data.get('indexes'):
        index_names = data['indexes']
    elif ('index' in data or 'indexes' in data) and not required:
        return []
    else:
        return None

    indexes = list(Index.select().where(Index.name << index_names))

    # Validate that all the index names exist.
    observed_names = set(index.name for index in indexes)
    invalid_names = []
    for index_name in index_names:
        if index_name not in observed_names:
            invalid_names.append(index_name)

    if invalid_names:
        error('The following indexes were not found: %s.' %
              ', '.join(invalid_names))

    return indexes

def _serialize_documents(document_query, include_score=False):
    # Eagerly load metadata and associated indexes.
    documents = prefetch(
        document_query,
        Metadata,
        IndexDocument,
        Index)
    document_list = []
    for document in documents:
        data = {
            'id': document.rowid,
            'identifier': document.identifier,
            'content': document.content}
        data['metadata'] = dict(
            (metadata.key, metadata.value)
            for metadata in document.metadata_set_prefetch)
        data['indexes'] = [
            idx_doc.index.name
            for idx_doc in document.indexdocument_set_prefetch]
        if include_score:
            data['score'] = document.score

        document_list.append(data)

    return document_list

def protect_view(fn):
    @wraps(fn)
    def inner(*args, **kwargs):
        if app.config['AUTHENTICATION']:
            # Check headers and request.args for `key=<key>`.
            api_key = None
            if request.headers.get('key'):
                api_key = request.headers['key']
            elif request.args.get('key'):
                api_key = request.args['key']
            if api_key != app.config['AUTHENTICATION']:
                return Response('Invalid API key'), 401
        return fn(*args, **kwargs)
    return inner

#
# Views.
#

@app.route('/', methods=['GET', 'POST'])
@protect_view
def index_list():
    """
    Main index for the SQLite search index. This view returns a JSON object
    containing a list of indexes (id and name) along with the number of
    documents stored in each index.

    This view can also be used to create new indexes by POSTing a `name`.
    """
    if request.method == 'POST':
        data = parse_post(['name'])

        with database.atomic():
            try:
                index = Index.create(name=data['name'])
            except IntegrityError:
                error('"%s" already exists.' % data['name'])

        return index_detail(index.name)

    query = (Index
             .select(Index, fn.COUNT(IndexDocument.id).alias('count'))
             .join(IndexDocument, JOIN_LEFT_OUTER)
             .group_by(Index)
             .order_by(Index.name))

    return jsonify({'indexes': [
        {'id': index.id, 'name': index.name, 'documents': index.count}
        for index in query
    ]})

@app.route('/<index_name>/', methods=['GET', 'POST', 'DELETE'])
@protect_view
def index_detail(index_name):
    """
    Detail view for an index. This view returns a JSON object with
    the index's id, name, and a paginated list of associated documents.

    Existing indexes can be renamed using this view by `POST`-ing a
    `name`, or deleted by issuing a `DELETE` request.
    """
    index = get_object_or_404(Index, Index.name == index_name)
    if request.method == 'POST':
        data = parse_post(['name'])
        index.name = data['name']
        with database.atomic():
            try:
                index.save()
            except IntegrityError:
                error('"%s" is already in use.' % index.name)
    elif request.method == 'DELETE':
        with database.atomic():
            (IndexDocument
             .delete()
             .where(IndexDocument.index == index)
             .execute())
            index.delete_instance()

        return jsonify({'success': True})

    pq = PaginatedQuery(
        index.documents,
        paginate_by=app.config['PAGINATE_BY'],
        page_var=app.config['PAGE_VAR'],
        check_bounds=False)

    return jsonify({
        'id': index.id,
        'name': index.name,
        'documents': _serialize_documents(pq.get_object_list()),
        'page': pq.get_page(),
        'pages': pq.get_page_count()})

@app.route('/documents/', methods=['GET', 'POST'])
@protect_view
def document_list():
    """
    Returns a paginated list of documents.

    Documents can be indexed by `POST`ing content, index(es) and,
    optionally, metadata.
    """
    if request.method == 'POST':
        data = parse_post(
            ['content'],
            ['identifier', 'index', 'indexes', 'metadata'])

        indexes = validate_indexes(data)
        if indexes is None:
            error('You must specify either an "index" or "indexes".')

        document = Document.create(
            content=data['content'],
            identifier=data.get('identifier'))

        if data.get('metadata'):
            document.metadata = data['metadata']

        for index in indexes:
            index.add_to_index(document)

        return jsonify({
            'id': document.rowid,
            'content': document.content,
            'indexes': [index.name for index in indexes],
            'metadata': document.metadata})

    query = Document.all()

    # Allow filtering by index.
    if request.args.get('index'):
        query = (query
                 .join(IndexDocument, JOIN_LEFT_OUTER)
                 .join(Index)
                 .where(Index.name == request.args['index']))

    pq = PaginatedQuery(
        query,
        paginate_by=app.config['PAGINATE_BY'],
        page_var=app.config['PAGE_VAR'],
        check_bounds=False)

    return jsonify({
        'documents': _serialize_documents(pq.get_object_list()),
        'page': pq.get_page(),
        'pages': pq.get_page_count()})

@app.route('/documents/<int:document_id>/', methods=['GET', 'POST', 'DELETE'])
@protect_view
def document_detail(document_id):
    """
    Return the details for an individual document. This view can also be
    used to update the `content`, `index(es)` and, optionally, `metadata`.
    To remove a document, issue a `DELETE` request to this view.
    """
    document = get_object_or_404(
        Document.all(),
        Document.rowid == document_id)
    return _document_detail(document)

@app.route('/documents/identifier/<identifier>/', methods=['GET', 'POST', 'DELETE'])
@protect_view
def document_by_identifier(identifier):
    document = get_object_or_404(
        Document.all(),
        Document.identifier == identifier)
    return _document_detail(document)

def _document_detail(document):
    if request.method == 'DELETE':
        with database.atomic():
            (IndexDocument
             .delete()
             .where(IndexDocument.document == document)
             .execute())
            Metadata.delete().where(Metadata.document == document).execute()
            document.delete_instance()
        return jsonify({'success': True})

    elif request.method == 'POST':
        data = parse_post([], [
            'content',
            'identifier',
            'index',
            'indexes',
            'metadata'])

        dirty = False
        for key in ('content', 'identifier'):
            if data.get(key):
                setattr(document, key, data[key])
                dirty = True

        if dirty:
            document.save()

        if 'metadata' in data:
            del document.metadata
            if data['metadata']:
                document.metadata = data['metadata']

        indexes = validate_indexes(data, required=False)
        if indexes is not None:
            with database.atomic():
                (IndexDocument
                 .delete()
                 .where(IndexDocument.document == document)
                 .execute())

                if indexes:
                    IndexDocument.insert_many([
                        {'index': index, 'document': document}
                        for index in indexes]).execute()
        else:
            indexes = document.get_indexes()

        index_names = [index.name for index in indexes]

    else:
        # GET requests.
        indexes = (Index
                   .select(Index.name)
                   .join(IndexDocument)
                   .where(IndexDocument.document == document))
        index_names = [index.name for index in indexes]

    return jsonify({
        'id': document.rowid,
        'identifier': document.identifier,
        'content': document.content,
        'indexes': index_names,
        'metadata': document.metadata,
    })

@app.route('/<index_name>/search/', methods=['GET'])
@protect_view
def search(index_name):
    """
    Search the index for documents matching the given query.
    """
    if not request.args.get('q'):
        error('Missing required search parameter "q".')

    search_query = request.args['q']
    ranking = request.args.get('ranking', Index.RANK_SIMPLE)
    if ranking and ranking not in (Index.RANK_SIMPLE, Index.RANK_BM25):
        error('Unrecognized "ranking" type.')

    filters = dict(
        (key, value) for key, value in request.args.items()
        if key not in ('page', 'q', 'key', 'ranking'))

    index = get_object_or_404(Index, Index.name == index_name)
    query = index.search(search_query, ranking, **filters)
    pq = PaginatedQuery(
        query,
        paginate_by=app.config['PAGINATE_BY'],
        page_var=app.config['PAGE_VAR'],
        check_bounds=False)

    return jsonify({
        'documents': _serialize_documents(
            pq.get_object_list(),
            include_score=ranking is not None),
        'page': pq.get_page(),
        'pages': pq.get_page_count()})


@app.errorhandler(InvalidRequestException)
def _handle_invalid_request(exc):
    return exc.response()

#
# Initialization, main().
#

def main():
    initialize_database(app.config['DATABASE'])
    if app.config['DEBUG']:
        app.run(host=app.config['HOST'], port=app.config['PORT'], debug=True)
    else:
        run_simple(
            hostname=app.config['HOST'],
            port=app.config['PORT'],
            application=app,
            threaded=True)

def initialize_database(database_file):
    database.init(database_file)

    with database.execution_context():
        Document.create_table(tokenize=app.config['STEM'], fail_silently=True)
        database.create_tables([Metadata, Index, IndexDocument], safe=True)

def panic(s, exit_code=1):
    sys.stderr.write('\033[91m%s\033[0m\n' % s)
    sys.stderr.flush()
    sys.exit(exit_code)

def get_option_parser():
    parser = optparse.OptionParser()
    parser.add_option(
        '-H',
        '--host',
        dest='host',
        help='The hostname to listen on. Defaults to 127.0.0.1.')
    parser.add_option(
        '-p',
        '--port',
        dest='port',
        help='The port to listen on. Defaults to 8000.',
        type='int')
    parser.add_option(
        '-s',
        '--stem',
        dest='stem',
        help='Specify stemming algorithm for content (default="porter").')
    parser.add_option(
        '-d',
        '--debug',
        action='store_true',
        dest='debug',
        help='Run Flask app in debug mode.')
    parser.add_option(
        '-c',
        '--config',
        dest='config',
        help='Configuration module (python file).')
    parser.add_option(
        '--paginate-by',
        dest='paginate_by',
        help='Number of documents displayed per page of results, default=50',
        type='int')
    parser.add_option(
        '-k',
        '--api-key',
        dest='api_key',
        help='Set the API key required to access Scout.')
    return parser


if __name__ == '__main__':
    option_parser = get_option_parser()
    options, args = option_parser.parse_args()

    if options.config:
        app.config.from_pyfile(options.config)

    if os.environ.get('SCOUT_CONFIG'):
        app.config.from_pyfile(os.environ['SCOUT_CONFIG'])

    if os.environ.get('SCOUT_DATABASE'):
        app.config['DATABASE'] = os.environ['SCOUT_DATABASE']

    if len(args) == 0 and not app.config.get('DATABASE'):
        panic('Error: missing required path to database file.')
    elif len(args) > 1:
        panic('Error: [%s] only accepts one argument, which is the path '
              'to the database file.' % __file__)
    elif args:
        app.config['DATABASE'] = args[0]

    # Handle command-line options. These values will override any values
    # that may have been specified in the config file.
    if options.api_key:
        app.config['AUTHENTICATION'] = options.api_key

    if options.debug:
        app.config['DEBUG'] = True

    if options.host:
        app.config['HOST'] = options.host

    if options.paginate_by:
        if options.paginate_by < 1 or options.paginate_by > 1000:
            panic('paginate-by must be between 1 and 1000')
        app.config['PAGINATE_BY'] = options.paginate_by

    if options.port:
        app.config['PORT'] = options.port

    if options.stem:
        if options.stem not in ('simple', 'porter'):
            panic('Unrecognized stemmer. Must be "porter" or "simple".')
        app.config['STEM'] = options.stem

    main()



import json
import optparse
import sys
import unittest
import urllib

from playhouse.test_utils import assert_query_count

from scout import app
from scout import database
from scout import Document
from scout import IndexDocument
from scout import Index
from scout import Metadata


def get_option_parser():
    parser = optparse.OptionParser()
    parser.add_option(
        '-q',
        '--quiet',
        action='store_true',
        dest='quiet')
    return parser


class BaseTestCase(unittest.TestCase):
    def setUp(self):
        database.drop_tables([
            Document,
            Metadata,
            Index,
            IndexDocument], safe=True)

        Document.create_table(tokenize='porter')
        database.create_tables([
            Metadata,
            Index,
            IndexDocument])

class TestModelAPIs(BaseTestCase):
    corpus = [
        ('A faith is a necessity to a man. Woe to him who believes in '
         'nothing.'),
        ('All who call on God in true faith, earnestly from the heart, '
         'will certainly be heard, and will receive what they have asked '
         'and desired.'),
        ('Be faithful in small things because it is in them that your '
         'strength lies.'),
        ('Faith consists in believing when it is beyond the power of '
         'reason to believe.'),
        ('Faith has to do with things that are not seen and hope with '
         'things that are not at hand.'),
    ]

    def setUp(self):
        super(TestModelAPIs, self).setUp()
        self.index = Index.create(name='default')

    def test_index_document(self):
        """
        Basic test case ensuring that content can be indexed and that the
        many-to-many relationship between documents and indexes is set
        up correctly.
        """
        content = 'huey is a sweet little kitty'
        doc = self.index.index(content=content)

        # Verify document properties.
        self.assertEqual(doc.rowid, 1)
        self.assertEqual(doc.content, content)
        self.assertEqual(doc.metadata, {})

        # Verify through relationship properties.
        self.assertEqual(IndexDocument.select().count(), 1)
        idx_doc = IndexDocument.get()
        self.assertEqual(idx_doc._data['document'], doc.rowid)
        self.assertEqual(idx_doc._data['index'], self.index.id)

    def test_index_with_metadata(self):
        """
        Test to ensure that content can be indexed with arbitrary key
        value metadata, which is stored as strings.
        """
        doc = self.index.index('test doc', foo='bar', nugget=33)
        self.assertEqual(doc.rowid, 1)
        self.assertEqual(doc.metadata, {'foo': 'bar', 'nugget': '33'})

    def test_reindex(self):
        """
        Test that an existing document can be re-indexed, updating the
        content and metadata in the process.
        """
        doc = self.index.index('test doc', foo='bar', baze='nug')
        doc_db = Document.select(Document.rowid, Document.content).get()
        self.assertTrue(doc_db.rowid is not None)
        self.assertEqual(doc_db.content, 'test doc')
        self.assertEqual(doc_db.metadata, {'foo': 'bar', 'baze': 'nug'})

        updated_doc = self.index.index(
            'updated doc',
            document=doc,
            foo='bazz',
            nug='x')
        self.assertEqual(Document.select().count(), 1)
        self.assertEqual(updated_doc.metadata, {'foo': 'bazz', 'nug': 'x'})

        u_doc_db = Document.select(Document.rowid, Document.content).get()
        self.assertEqual(u_doc_db.content, 'updated doc')
        self.assertEqual(u_doc_db.rowid, doc_db.rowid)
        self.assertEqual(u_doc_db.metadata, {'foo': 'bazz', 'nug': 'x'})

        # Verify through relationship properties.
        self.assertEqual(IndexDocument.select().count(), 1)
        idx_doc = IndexDocument.get()
        self.assertEqual(idx_doc._data['document'], u_doc_db.rowid)
        self.assertEqual(idx_doc._data['index'], self.index.id)

    def test_multi_index(self):
        """
        Test that documents can be stored in multiple indexes.
        """
        self.index.delete_instance()

        indexes = [Index.create(name='idx-%s' % i) for i in range(3)]
        document = Document.create(content='hueybear')
        for index in indexes:
            index.index(
                document.content,
                document)

        self.assertEqual(Document.select().count(), 1)
        self.assertEqual(Index.select().count(), 3)
        self.assertEqual(IndexDocument.select().count(), 3)
        query = (IndexDocument
                 .select(Index.name, IndexDocument.document)
                 .join(Index)
                 .order_by(Index.name)
                 .dicts())
        idx_doc_data = [idx_doc for idx_doc in query]
        self.assertEqual(idx_doc_data, [
            {'document': document.rowid, 'name': 'idx-0'},
            {'document': document.rowid, 'name': 'idx-1'},
            {'document': document.rowid, 'name': 'idx-2'},
        ])

    def test_search(self):
        """
        Basic tests for simple string searches of a single index. Use both
        the simple and bm25 ranking algorithms.
        """
        for idx, content in enumerate(self.corpus):
            self.index.index(content=content)

        def assertSearch(phrase, indexes, ranking=Index.RANK_SIMPLE):
            results = [doc.content
                       for doc in self.index.search(phrase, ranking)]
            self.assertEqual(results, [self.corpus[idx] for idx in indexes])

        assertSearch('believe', [3, 0])
        assertSearch('faith man', [0])
        assertSearch('faith thing', [4, 2])
        assertSearch('things', [4, 2])
        assertSearch('blah', [])
        assertSearch('', [])

        assertSearch('believe', [3, 0], Index.RANK_BM25)  # Same result.
        assertSearch('faith thing', [2, 4], Index.RANK_BM25)  # Swapped.
        assertSearch('things', [4, 2], Index.RANK_BM25)  # Same result.
        assertSearch('blah', [], Index.RANK_BM25)  # No results, works.
        assertSearch('', [], Index.RANK_BM25)


class TestSearchViews(BaseTestCase):
    def setUp(self):
        super(TestSearchViews, self).setUp()
        self.app = app.test_client()
        app.config['AUTHENTICATION'] = None

    def post_json(self, url, data, parse_response=True):
        response = self.app.post(url, data=json.dumps(data))
        if parse_response:
            return json.loads(response.data)
        return response

    def test_create_index(self):
        data = self.post_json('/', {'name': 'TestIndex'})
        self.assertEqual(data['name'], 'TestIndex')
        self.assertEqual(data['documents'], [])
        self.assertEqual(Index.select().count(), 1)

    def test_create_missing_name(self):
        error = {'error': 'Missing required fields: name'}
        data = self.post_json('/', {})
        self.assertEqual(data, error)

        data = self.post_json('/', {'name': None})
        self.assertEqual(data, error)

    def test_create_invalid_json(self):
        response = self.app.post('/', data='not json')
        data = json.loads(response.data)
        self.assertEqual(
            data,
            {'error': 'Unable to parse JSON data from request.'})

    def test_index_list(self):
        for i in range(3):
            Index.create(name='i%s' % i)

        response = self.app.get('/')
        data = json.loads(response.data)
        self.assertEqual(data['indexes'], [
            {'documents': 0, 'id': 1, 'name': 'i0'},
            {'documents': 0, 'id': 2, 'name': 'i1'},
            {'documents': 0, 'id': 3, 'name': 'i2'},
        ])

    def test_index_missing(self):
        response = self.app.get('/missing/')
        self.assertEqual(response.status_code, 404)

    def test_index_detail(self):
        idx_a = Index.create(name='idx-a')
        idx_b = Index.create(name='idx-b')
        for i in range(11):
            idx_a.index('document-%s' % i, foo='bar-%s' % i)

        b_doc = idx_b.index('both-doc')
        idx_a.index(b_doc.content, b_doc)

        response = self.app.get('/idx-a/')
        data = json.loads(response.data)
        self.assertEqual(data['page'], 1)
        self.assertEqual(data['pages'], 2)
        self.assertEqual(len(data['documents']), 10)
        doc = data['documents'][0]
        self.assertEqual(doc, {
            'content': 'document-0',
            'id': 1,
            'identifier': None,
            'indexes': ['idx-a'],
            'metadata': {'foo': 'bar-0'}})

        response = self.app.get('/idx-a/?page=2')
        data = json.loads(response.data)
        self.assertEqual(data['page'], 2)
        self.assertEqual(data['pages'], 2)
        self.assertEqual(len(data['documents']), 2)

        response = self.app.get('/idx-b/')
        data = json.loads(response.data)
        self.assertEqual(data['page'], 1)
        self.assertEqual(data['pages'], 1)
        self.assertEqual(len(data['documents']), 1)
        doc = data['documents'][0]
        self.assertEqual(doc, {
            'content': 'both-doc',
            'id': 12,
            'identifier': None,
            'indexes': ['idx-b', 'idx-a'],
            'metadata': {}})

    def test_index_update_delete(self):
        idx = Index.create(name='idx')
        alt_idx = Index.create(name='alt-idx')
        doc = idx.index(content='foo')
        alt_idx.index(doc.content, doc)
        idx.index('idx only')
        alt_idx.index('alt only')

        response = self.post_json('/idx/', {'name': 'idx-updated'})
        self.assertEqual(response['id'], idx.id)
        self.assertEqual(response['name'], 'idx-updated')
        self.assertEqual(
            [doc['content'] for doc in response['documents']],
            ['foo', 'idx only'])

        response = self.app.delete('/idx-updated/')
        data = json.loads(response.data)
        self.assertEqual(data, {'success': True})

        self.assertEqual(Document.select().count(), 3)
        self.assertEqual(IndexDocument.select().count(), 2)
        self.assertEqual(Index.select().count(), 1)

    def test_index_document(self):
        idx_a = Index.create(name='idx-a')
        idx_b = Index.create(name='idx-b')
        response = self.post_json('/documents/', {
            'content': 'doc 1',
            'index': 'idx-a',
            'metadata': {'k1': 'v1', 'k2': 'v2'}})

        self.assertEqual(response, {
            'content': 'doc 1',
            'id': 1,
            'indexes': ['idx-a'],
            'metadata': {'k1': 'v1', 'k2': 'v2'}})

        response = self.post_json('/documents/', {
            'content': 'doc 2',
            'indexes': ['idx-a', 'idx-b']})
        self.assertEqual(response, {
            'content': 'doc 2',
            'id': 2,
            'indexes': ['idx-a', 'idx-b'],
            'metadata': {}})

    def test_index_document_validation(self):
        idx = Index.create(name='idx')
        response = self.post_json('/documents/', {'content': 'foo'})
        self.assertEqual(
            response['error'],
            'You must specify either an "index" or "indexes".')

        response = self.post_json('/documents/', {'content': 'x', 'index': ''})
        self.assertEqual(
            response['error'],
            'You must specify either an "index" or "indexes".')

        response = self.post_json('/documents/', {
            'content': 'foo',
            'index': 'missing'})
        self.assertEqual(
            response['error'],
            'The following indexes were not found: missing.')

        response = self.post_json('/documents/', {
            'content': 'foo',
            'indexes': ['missing', 'idx', 'blah']})
        self.assertEqual(
            response['error'],
            'The following indexes were not found: missing, blah.')
        self.assertEqual(Document.select().count(), 0)

    def test_document_detail_get(self):
        idx = Index.create(name='idx')
        doc = idx.index('test doc', foo='bar')
        alt_doc = idx.index('alt doc')

        response = self.app.get('/documents/%s/' % doc.rowid)
        data = json.loads(response.data)
        self.assertEqual(data, {
            'content': 'test doc',
            'id': doc.rowid,
            'identifier': None,
            'indexes': ['idx'],
            'metadata': {'foo': 'bar'}})

    def test_document_detail_by_identifier(self):
        idx = Index.create(name='idx')
        doc = idx.index('test doc', identifier='td', foo='bar')
        alt_doc = idx.index('alt doc', identifier='ad')

        response = self.app.get('/documents/identifier/td/')
        data = json.loads(response.data)
        self.assertEqual(data, {
            'content': 'test doc',
            'id': doc.rowid,
            'identifier': doc.identifier,
            'indexes': ['idx'],
            'metadata': {'foo': 'bar'}})

    def refresh_doc(self, doc):
        return Document.all().where(Document.rowid == doc.rowid).get()

    def test_document_detail_post(self):
        idx = Index.create(name='idx')
        alt_idx = Index.create(name='alt-idx')
        doc = idx.index('test doc', foo='bar', nug='baze')
        alt_doc = idx.index('alt doc')

        url = '/documents/%s/' % doc.rowid

        def assertDoc(doc, content, metadata=None, indexes=None):
            doc_db = self.refresh_doc(doc)
            self.assertEqual(doc_db.content, content)
            self.assertEqual(
                [idx.name for idx in doc_db.get_indexes()],
                indexes or [])
            self.assertEqual(doc_db.metadata, metadata or {})

        # Update the content.
        response = self.post_json(url, {'content': 'updated'})
        assertDoc(doc, 'updated', {'foo': 'bar', 'nug': 'baze'}, ['idx'])

        # Test updating metadata.
        response = self.post_json(url, {'metadata': dict(
            doc.metadata, nug='baz', herp='derp')})
        assertDoc(
            doc,
            'updated',
            {'foo': 'bar', 'nug': 'baz', 'herp': 'derp'},
            ['idx'])

        # Test clearing metadata.
        response = self.post_json(url, {'metadata': None})
        assertDoc(doc, 'updated', {}, ['idx'])

        # Test updating indexes.
        response = self.post_json(url, {'indexes': ['idx', 'alt-idx']})
        assertDoc(doc, 'updated', {}, ['alt-idx', 'idx'])

        # Test clearing indexes.
        response = self.post_json(url, {'indexes': []})
        assertDoc(doc, 'updated', {}, [])

        # Ensure alt_doc has not been affected.
        assertDoc(alt_doc, 'alt doc', {}, ['idx'])

        # Sanity check.
        self.assertEqual(Document.select().count(), 2)

    def test_document_detail_delete(self):
        idx = Index.create(name='idx')
        alt_idx = Index.create(name='alt-idx')

        d1 = idx.index('doc 1', k1='v1', k2='v2')
        d2 = idx.index('doc 2', k3='v3')
        alt_idx.add_to_index(d1)
        alt_idx.add_to_index(d2)

        self.assertEqual(Metadata.select().count(), 3)

        response = self.app.delete('/documents/%s/' % d2.rowid)
        data = json.loads(response.data)
        self.assertEqual(data, {'success': True})

        self.assertEqual(Metadata.select().count(), 2)

        response = self.app.delete('/documents/%s/' % d2.rowid)
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Document.select().count(), 1)
        self.assertEqual(IndexDocument.select().count(), 2)
        self.assertEqual([d.rowid for d in idx.documents], [d1.rowid])
        self.assertEqual([d.rowid for d in alt_idx.documents], [d1.rowid])

    def search(self, index, query, page=1, **filters):
        params = urllib.urlencode(dict(filters, q=query, page=page))
        response = self.app.get('/%s/search/?%s' % (index, params))
        return json.loads(response.data)

    def test_search(self):
        idx = Index.create(name='idx')
        phrases = ['foo', 'bar', 'baz', 'nug nugs', 'blah nuggie foo', 'huey',
                   'zaizee']
        for phrase in phrases:
            idx.index('document %s' % phrase, special=True)

        for i in range(10):
            idx.index('document %s' % i, special=False)

        response = self.search('idx', 'docum*')
        self.assertEqual(response['page'], 1)
        self.assertEqual(response['pages'], 2)
        self.assertEqual(len(response['documents']), 10)

        first_doc = response['documents'][0]
        self.assertEqual(round(first_doc['score'], 3), .059)

        response = self.search('idx', 'document', 2)
        self.assertEqual(len(response['documents']), 7)

        response = self.search('idx', 'doc* nug*')
        self.assertEqual(response['page'], 1)
        self.assertEqual(response['pages'], 1)
        self.assertEqual(len(response['documents']), 2)
        doc1, doc2 = response['documents']

        self.assertEqual(doc1, {
            'content': 'document nug nugs',
            'id': doc1['id'],
            'identifier': None,
            'indexes': ['idx'],
            'metadata': {'special': 'True'},
            'score': doc1['score']})
        self.assertEqual(round(doc1['score'], 4), .7255)

        self.assertEqual(doc2, {
            'content': 'document blah nuggie foo',
            'id': doc2['id'],
            'identifier': None,
            'indexes': ['idx'],
            'metadata': {'special': 'True'},
            'score': doc2['score']})
        self.assertEqual(round(doc2['score'], 4), .3922)

        response = self.search('idx', 'missing')
        self.assertEqual(len(response['documents']), 0)

        response = self.search('idx', 'nug', ranking='bm25')
        doc = response['documents'][0]
        self.assertEqual(doc['content'], 'document nug nugs')
        self.assertEqual(round(doc['score'], 3), 2.891)

    def test_search_filters(self):
        idx = Index.create(name='idx')
        data = (
            ('huey document', {'name': 'huey', 'kitty': 'yes'}),
            ('zaizee document', {'name': 'zaizee', 'kitty': 'yes'}),
            ('little huey bear', {'name': 'huey', 'kitty': 'yes'}),
            ('uncle huey', {'kitty': 'no'}),
            ('michael nuggie document', {'name': 'mickey', 'kitty': 'no'}),
        )
        for content, metadata in data:
            idx.index(content, **metadata)

        def assertResults(query, metadata, expected):
            results = self.search('idx', query, **metadata)
            content = [document['content']
                       for document in results['documents']]
            self.assertEqual(content, expected)

        assertResults(
            'huey',
            {},
            ['huey document', 'little huey bear', 'uncle huey'])
        assertResults(
            'huey',
            {'kitty': 'yes'},
            ['huey document', 'little huey bear'])
        assertResults(
            'huey',
            {'kitty': 'yes', 'name': 'huey'},
            ['huey document', 'little huey bear'])
        assertResults(
            'docu*',
            {'kitty': 'yes'},
            ['huey document', 'zaizee document'])

    def test_query_count(self):
        idx_a = Index.create(name='idx-a')
        idx_b = Index.create(name='idx-b')
        phrases = ['foo', 'bar', 'baze', 'nug', 'nuggie']
        for phrase in phrases:
            phrase = 'document ' + phrase
            doc = idx_a.index(phrase)
            idx_b.index(phrase, doc, foo='bar', baze='nug')

        for idx in ['idx-a', 'idx-b']:
            for query in ['nug', 'nug*', 'document', 'missing']:
                with assert_query_count(6):
                    # 1. Get index.
                    # 2. Prefetch indexes.
                    # 3. Prefetch index documents.
                    # 4. Prefetch metadata
                    # 5. Fetch documents (top of prefetch).
                    # 6. COUNT(*) for pagination.
                    self.search(idx, query)

                with assert_query_count(6):
                    self.search(idx, query, foo='bar')

        with assert_query_count(6):
            # Same as above.
            self.app.get('/idx-a/')

        with assert_query_count(5):
            # Same as above minus first query for index.
            self.app.get('/documents/')

        for i in range(10):
            Index.create(name='idx-%s' % i)

        with assert_query_count(1):
            self.app.get('/')

    def test_authentication(self):
        Index.create(name='idx')

        app.config['AUTHENTICATION'] = 'test'
        resp = self.app.get('/')
        self.assertEqual(resp.status_code, 401)
        self.assertEqual(resp.data, 'Invalid API key')

        resp = self.app.get('/?key=tesss')
        self.assertEqual(resp.status_code, 401)

        resp = self.app.get('/', headers={'key': 'tesss'})
        self.assertEqual(resp.status_code, 401)

        resp = self.app.get('/?key=test')
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(json.loads(resp.data), {'indexes': [
            {'id': 1, 'name': 'idx', 'documents': 0}]})

        resp = self.app.get('/', headers={'key': 'test'})
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(json.loads(resp.data), {'indexes': [
            {'id': 1, 'name': 'idx', 'documents': 0}]})


def main():
    option_parser = get_option_parser()
    options, args = option_parser.parse_args()
    database.init(':memory:')
    app.config['PAGINATE_BY'] = 10
    unittest.main(argv=sys.argv, verbosity=not options.quiet and 2 or 0)


if __name__ == '__main__':
    main()
